==> hlog-0.16-rc1.tar.gz: hlog-0.16-rc1/daemon/wsjtx.c <==
/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/buffer.h>
#include <jeffpc/time.h>

#include <hlog/time.h>
#include <hlog/qso.h>

#include "udpsvc.h"
#include "wsjtx.h"

#define WSJTX_HOST NULL
#define WSJTX_PORT 2237

#define WSJTX_MAGIC 0xadbccbda
#define WSJTX_SCHEMA 2 /* Qt 5.2 */

struct rawmessage {
	uint32_t magic;
	uint32_t schema;
	uint32_t type;
	uint8_t payload[];
};

struct rstr {
	const char *ptr;
	size_t len;
};

struct message {
	uint32_t magic;
	uint32_t schema;
	uint32_t type;

	union {
		struct {
			struct rstr id;
			struct time timeoff;
			struct rstr dx_call;
			struct rstr dx_grid;
			uint64_t tx_freq;
			struct rstr mode;
			struct rstr rsts;
			struct rstr rstr;
			struct rstr tx_power;
			struct rstr comment;
			struct rstr name;
			struct time timeon;
			struct rstr op_call;
			struct rstr my_call;
			struct rstr my_grid;
			struct rstr exch_sent;
			struct rstr exch_recv;
		} t5; /* QSO Logged (Out) */
	};
};

static bool decode_u8(uint8_t *ptr, struct buffer *buf)
{
	uint8_t val;

	if (buffer_read(buf, &val, sizeof(val)) != sizeof(val))
		return false;

	*ptr = be8_to_cpu(val);

	return true;
}

static bool decode_u32(uint32_t *ptr, struct buffer *buf)
{
	uint32_t val;

	if (buffer_read(buf, &val, sizeof(val)) != sizeof(val))
		return false;

	*ptr = be32_to_cpu(val);

	return true;
}

static bool decode_u64(uint64_t *ptr, struct buffer *buf)
{
	uint64_t val;

	if (buffer_read(buf, &val, sizeof(val)) != sizeof(val))
		return false;

	*ptr = be64_to_cpu(val);

	return true;
}

static bool decode_str(struct rstr *ptr, struct buffer *buf)
{
	uint32_t len;

	if (!decode_u32(&len, buf))
		return false;

	ptr->ptr = len ? buffer_data_current(buf) : NULL;
	ptr->len = len;

	if (buffer_seek(buf, len, SEEK_CUR) < 0)
		return false;

	return true;
}

static bool decode_datetime(struct time *ptr, struct buffer *buf)
{
	uint64_t daynum;
	uint32_t msecs;
	uint8_t spec;
	struct tm tm;
	time_t time;

	if (!decode_u64(&daynum, buf))
		return false;
	if (!decode_u32(&msecs, buf))
		return false;
	if (!decode_u8(&spec, buf))
		return false;

	if (spec != 1)
		return false; /* not UTC */

	/* convert <daynum, msecs> days to unix timestamp */
	time = ((daynum - 2440588) * 86400) + (msecs / 1000);

	if (!gmtime_r(&time, &tm))
		return false;

	ptr->have = TIME_PART_DATE | TIME_PART_TIME_HMS;
	ptr->year = tm.tm_year + 1900;
	ptr->month = tm.tm_mon + 1;
	ptr->day = tm.tm_mday;
	ptr->hour = tm.tm_hour;
	ptr->min = tm.tm_min;
	ptr->sec = tm.tm_sec;

	return true;
}

static bool decode_qso_logged(struct message *msg, struct buffer *buf)
{
	if (!decode_str(&msg->t5.id, buf))
		return false;
	if (!decode_datetime(&msg->t5.timeoff, buf))
		return false;
	if (!decode_str(&msg->t5.dx_call, buf))
		return false;
	if (!decode_str(&msg->t5.dx_grid, buf))
		return false;
	if (!decode_u64(&msg->t5.tx_freq, buf))
		return false;
	if (!decode_str(&msg->t5.mode, buf))
		return false;
	if (!decode_str(&msg->t5.rsts, buf))
		return false;
	if (!decode_str(&msg->t5.rstr, buf))
		return false;
	if (!decode_str(&msg->t5.tx_power, buf))
		return false;
	if (!decode_str(&msg->t5.comment, buf))
		return false;
	if (!decode_str(&msg->t5.name, buf))
		return false;
	if (!decode_datetime(&msg->t5.timeon, buf))
		return false;
	if (!decode_str(&msg->t5.op_call, buf))
		return false;
	if (!decode_str(&msg->t5.my_call, buf))
		return false;
	if (!decode_str(&msg->t5.my_grid, buf))
		return false;
	if (!decode_str(&msg->t5.exch_sent, buf))
		return false;
	if (!decode_str(&msg->t5.exch_recv, buf))
		return false;
	return true;
}

static char *terminate_str(char *out, size_t outsize,
			   const char *in, size_t insize)
{
	ASSERT3U(outsize, >=, (insize + 1));

	memcpy(out, in, insize);

	out[insize] = '\0';

	return out;
}

static int log_qso(struct message *msg, const void *raw, size_t rawlen,
		   int data_dir_fd)
{
	char tmp[1024];
	struct qso *qso;
	int ret;

	/*
	 * Set up a struct qso with the received information.
	 */
	qso = qso_alloc();
	if (IS_ERR(qso))
		return PTR_ERR(qso);

	/* QSO */
	qso->start = msg->t5.timeon;
	qso->end = msg->t5.timeoff;

	if (msg->t5.comment.len) {
		ret = qso_set_comment(qso, STR_DUP_LEN(msg->t5.comment.ptr,
						       msg->t5.comment.len));
		if (ret)
			goto err;
	}

	/* QSO side - tx */
	ret = qso_set_station_call(&qso->tx,
				   terminate_str(tmp, sizeof(tmp),
						 msg->t5.my_call.ptr,
						 msg->t5.my_call.len));
	if (ret)
		goto err;
	ret = qso_set_operator_call(&qso->tx,
				    terminate_str(tmp, sizeof(tmp),
						  msg->t5.op_call.ptr,
						  msg->t5.op_call.len));
	if (ret)
		goto err;
	ASSERT3U(msg->t5.tx_freq, !=, 0);
	ret = qso_set_freq_val(&qso->tx, VAL_ALLOC_INT(msg->t5.tx_freq));
	if (ret)
		goto err;
	ret = qso_set_grid(&qso->tx,
			   terminate_str(tmp, sizeof(tmp),
					 msg->t5.my_grid.ptr,
					 msg->t5.my_grid.len));
	if (ret)
		goto err;
	ret = qso_set_power_user(&qso->tx,
				 terminate_str(tmp, sizeof(tmp),
					       msg->t5.tx_power.ptr,
					       msg->t5.tx_power.len));
	if (ret)
		goto err;
	ret = qso_set_mode(&qso->tx,
			   terminate_str(tmp, sizeof(tmp),
					 msg->t5.mode.ptr,
					 msg->t5.mode.len));
	if (ret)
		goto err;
	ret = qso_set_rst(&qso->tx,
			  terminate_str(tmp, sizeof(tmp),
					msg->t5.rsts.ptr,
					msg->t5.rsts.len));
	if (ret)
		goto err;
	/* TODO: set exchange sent */

	/* QSO side - rx */
	ret = qso_set_station_call(&qso->rx,
				   terminate_str(tmp, sizeof(tmp),
						 msg->t5.dx_call.ptr,
						 msg->t5.dx_call.len));
	if (ret)
		goto err;
	if (msg->t5.dx_grid.len) {
		ret = qso_set_grid(&qso->rx,
				   terminate_str(tmp, sizeof(tmp),
						 msg->t5.dx_grid.ptr,
						 msg->t5.dx_grid.len));
		if (ret)
			goto err;
	}
	ret = qso_set_mode_val(&qso->rx, qso_get_mode_val(&qso->tx));
	if (ret)
		goto err;
	ret = qso_set_rst(&qso->rx,
			  terminate_str(tmp, sizeof(tmp),
					msg->t5.rstr.ptr,
					msg->t5.rstr.len));
	if (ret)
		goto err;
	if (msg->t5.name.len) {
		ret = qso_set_name(&qso->rx, STR_DUP_LEN(msg->t5.name.ptr,
							 msg->t5.name.len));
		if (ret)
			goto err;
	}
	/* TODO: set exchange received */

	/* stash the current time & raw message in an additional field */
	ret = nvl_set_int(qso->additional, "wsjtx-import-time", gettime());
	if (ret)
		goto err;

	ret = nvl_set(qso->additional, "wsjtx-raw-udp",
		      VAL_ALLOC_BLOB_STATIC(raw, rawlen));
	if (ret)
		goto err;

	/*
	 * Now, save the QSO.
	 */
	ret = qso_save(qso, data_dir_fd);
	if (!ret) {
		struct str *mode, *call, *grid;

		mode = qso_get_mode(&qso->tx);
		call = qso_get_station_call(&qso->rx);
		grid = qso_get_grid(&qso->rx);

		cmn_err(CE_INFO, "wsjtx: saved %s %s %s %s",
			qso->uuid_str, str_cstr(mode), str_cstr(call),
			str_cstr(grid));

		str_putref(mode);
		str_putref(call);
		str_putref(grid);
	}

err:
	qso_free(qso);

	return ret;
}

static void wsjtx_process(void *buf, size_t len, void *_data_dir_fd)
{
	const int data_dir_fd = (intptr_t) _data_dir_fd;
	struct rawmessage *raw = buf;
	struct message msg;
	struct buffer payload;
	int ret;

	if (len < (3 * sizeof(uint32_t))) {
		cmn_err(CE_INFO, "Message is too short (got %zu bytes, "
			"need at least %zu)", len, 3 * sizeof(uint32_t));
		return;
	}

	/*
	 * Handle the header and message type
	 */
	msg.magic = be32_to_cpu(raw->magic);
	msg.schema = be32_to_cpu(raw->schema);
	msg.type = be32_to_cpu(raw->type);

	if (msg.magic != WSJTX_MAGIC) {
		cmn_err(CE_INFO, "Magic mismatch (got %08x, expected %08x)",
			msg.magic, WSJTX_MAGIC);
		return;
	}

	if (msg.schema != WSJTX_SCHEMA) {
		cmn_err(CE_INFO, "Unsupported schema (got %u, expected %u)",
			msg.schema, WSJTX_SCHEMA);
		return;
	}

	buffer_init_static(&payload, &raw->payload,
			   len - offsetof(struct rawmessage, payload),
			   len - offsetof(struct rawmessage, payload),
			   false);

	switch (msg.type) {
		case 5: /* QSO Logged */
			if (!decode_qso_logged(&msg, &payload))
				return;

			ret = log_qso(&msg, buf, len, data_dir_fd);
			if (ret)
				cmn_err(CE_WARN, "Failed to log QSO: %s",
					xstrerror(ret));
			break;
		default:
			/* ignore */
			break;
	}
}

int wsjtx_start(int data_dir_fd)
{
	int ret;

	ret = udpsvc(WSJTX_HOST, WSJTX_PORT, wsjtx_process,
		     (void *)(intptr_t) data_dir_fd);
	if (ret)
		cmn_err(CE_ERROR, "wsjtx: failed to start listener: %s",
			xstrerror(ret));

	return ret;
}

==> hlog-0.17.tar.gz: hlog-0.17/daemon/wsjtx.c <==
/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/buffer.h>
#include <jeffpc/time.h>

#include <hlog/time.h>
#include <hlog/qso.h>

#include "announce.h"
#include "udpsvc.h"
#include "wsjtx.h"

#define WSJTX_HOST NULL
#define WSJTX_PORT 2237

#define WSJTX_MAGIC 0xadbccbda
#define WSJTX_SCHEMA 2 /* Qt 5.2 */

struct rawmessage {
	uint32_t magic;
	uint32_t schema;
	uint32_t type;
	uint8_t payload[];
};

struct rstr {
	const char *ptr;
	size_t len;
};

struct message {
	uint32_t magic;
	uint32_t schema;
	uint32_t type;

	union {
		struct {
			struct rstr id;
			struct time timeoff;
			struct rstr dx_call;
			struct rstr dx_grid;
			uint64_t tx_freq;
			struct rstr mode;
			struct rstr rsts;
			struct rstr rstr;
			struct rstr tx_power;
			struct rstr comment;
			struct rstr name;
			struct time timeon;
			struct rstr op_call;
			struct rstr my_call;
			struct rstr my_grid;
			struct rstr exch_sent;
			struct rstr exch_recv;
		} t5; /* QSO Logged (Out) */
	};
};

static bool decode_u8(uint8_t *ptr, struct buffer *buf)
{
	uint8_t val;

	if (buffer_read(buf, &val, sizeof(val)) != sizeof(val))
		return false;

	*ptr = be8_to_cpu(val);

	return true;
}

static bool decode_u32(uint32_t *ptr, struct buffer *buf)
{
	uint32_t val;

	if (buffer_read(buf, &val, sizeof(val)) != sizeof(val))
		return false;

	*ptr = be32_to_cpu(val);

	return true;
}

static bool decode_u64(uint64_t *ptr, struct buffer *buf)
{
	uint64_t val;

	if (buffer_read(buf, &val, sizeof(val)) != sizeof(val))
		return false;

	*ptr = be64_to_cpu(val);

	return true;
}

static bool decode_str(struct rstr *ptr, struct buffer *buf)
{
	uint32_t len;

	if (!decode_u32(&len, buf))
		return false;

	ptr->ptr = len ? buffer_data_current(buf) : NULL;
	ptr->len = len;

	if (buffer_seek(buf, len, SEEK_CUR) < 0)
		return false;

	return true;
}

static bool decode_datetime(struct time *ptr, struct buffer *buf)
{
	uint64_t daynum;
	uint32_t msecs;
	uint8_t spec;
	struct tm tm;
	time_t time;

	if (!decode_u64(&daynum, buf))
		return false;
	if (!decode_u32(&msecs, buf))
		return false;
	if (!decode_u8(&spec, buf))
		return false;

	if (spec != 1)
		return false; /* not UTC */

	/* convert <daynum, msecs> days to unix timestamp */
	time = ((daynum - 2440588) * 86400) + (msecs / 1000);

	if (!gmtime_r(&time, &tm))
		return false;

	ptr->have = TIME_PART_DATE | TIME_PART_TIME_HMS;
	ptr->year = tm.tm_year + 1900;
	ptr->month = tm.tm_mon + 1;
	ptr->day = tm.tm_mday;
	ptr->hour = tm.tm_hour;
	ptr->min = tm.tm_min;
	ptr->sec = tm.tm_sec;

	return true;
}

static bool decode_qso_logged(struct message *msg, struct buffer *buf)
{
	if (!decode_str(&msg->t5.id, buf))
		return false;
	if (!decode_datetime(&msg->t5.timeoff, buf))
		return false;
	if (!decode_str(&msg->t5.dx_call, buf))
		return false;
	if (!decode_str(&msg->t5.dx_grid, buf))
		return false;
	if (!decode_u64(&msg->t5.tx_freq, buf))
		return false;
	if (!decode_str(&msg->t5.mode, buf))
		return false;
	if (!decode_str(&msg->t5.rsts, buf))
		return false;
	if (!decode_str(&msg->t5.rstr, buf))
		return false;
	if (!decode_str(&msg->t5.tx_power, buf))
		return false;
	if (!decode_str(&msg->t5.comment, buf))
		return false;
	if (!decode_str(&msg->t5.name, buf))
		return false;
	if (!decode_datetime(&msg->t5.timeon, buf))
		return false;
	if (!decode_str(&msg->t5.op_call, buf))
		return false;
	if (!decode_str(&msg->t5.my_call, buf))
		return false;
	if (!decode_str(&msg->t5.my_grid, buf))
		return false;
	if (!decode_str(&msg->t5.exch_sent, buf))
		return false;
	if (!decode_str(&msg->t5.exch_recv, buf))
		return false;
	return true;
}

static char *terminate_str(char *out, size_t outsize,
			   const char *in, size_t insize)
{
	ASSERT3U(outsize, >=, (insize + 1));

	memcpy(out, in, insize);

	out[insize] = '\0';

	return out;
}

static int log_qso(struct message *msg, const void *raw, size_t rawlen)
{
	char tmp[1024];
	struct qso *qso;
	int ret;

	/*
	 * Set up a struct qso with the received information.
	 */
	qso = qso_alloc();
	if (IS_ERR(qso))
		return PTR_ERR(qso);

	/* QSO */
	qso->start = msg->t5.timeon;
	qso->end = msg->t5.timeoff;

	if (msg->t5.comment.len) {
		ret = qso_set_comment_str(qso,
					  STR_DUP_LEN(msg->t5.comment.ptr,
						      msg->t5.comment.len));
		if (ret)
			goto err;
	}

	/* QSO side - tx */
	ret = qso_set_station_call(&qso->tx,
				   terminate_str(tmp, sizeof(tmp),
						 msg->t5.my_call.ptr,
						 msg->t5.my_call.len));
	if (ret)
		goto err;
	ret = qso_set_operator_call(&qso->tx,
				    terminate_str(tmp, sizeof(tmp),
						  msg->t5.op_call.ptr,
						  msg->t5.op_call.len));
	if (ret)
		goto err;
	ASSERT3U(msg->t5.tx_freq, !=, 0);
	ret = qso_set_freq_val(&qso->tx, VAL_ALLOC_INT(msg->t5.tx_freq));
	if (ret)
		goto err;
	ret = qso_set_grid(&qso->tx,
			   terminate_str(tmp, sizeof(tmp),
					 msg->t5.my_grid.ptr,
					 msg->t5.my_grid.len));
	if (ret)
		goto err;
	ret = qso_set_power_user(&qso->tx,
				 terminate_str(tmp, sizeof(tmp),
					       msg->t5.tx_power.ptr,
					       msg->t5.tx_power.len));
	if (ret)
		goto err;
	ret = qso_set_mode(&qso->tx,
			   terminate_str(tmp, sizeof(tmp),
					 msg->t5.mode.ptr,
					 msg->t5.mode.len));
	if (ret)
		goto err;
	ret = qso_set_rst(&qso->tx,
			  terminate_str(tmp, sizeof(tmp),
					msg->t5.rsts.ptr,
					msg->t5.rsts.len));
	if (ret)
		goto err;
	/* TODO: set exchange sent */

	/* QSO side - rx */
	ret = qso_set_station_call(&qso->rx,
				   terminate_str(tmp, sizeof(tmp),
						 msg->t5.dx_call.ptr,
						 msg->t5.dx_call.len));
	if (ret)
		goto err;
	if (msg->t5.dx_grid.len) {
		ret = qso_set_grid(&qso->rx,
				   terminate_str(tmp, sizeof(tmp),
						 msg->t5.dx_grid.ptr,
						 msg->t5.dx_grid.len));
		if (ret)
			goto err;
	}
	ret = qso_set_mode_val(&qso->rx, qso_get_mode_val(&qso->tx));
	if (ret)
		goto err;
	ret = qso_set_rst(&qso->rx,
			  terminate_str(tmp, sizeof(tmp),
					msg->t5.rstr.ptr,
					msg->t5.rstr.len));
	if (ret)
		goto err;
	if (msg->t5.name.len) {
		ret = qso_set_name_str(&qso->rx,
				       STR_DUP_LEN(msg->t5.name.ptr,
						   msg->t5.name.len));
		if (ret)
			goto err;
	}
	/* TODO: set exchange received */

	/* stash the current time & raw message in an additional field */
	ret = nvl_set_int(qso->additional, "wsjtx-import-time", gettime());
	if (ret)
		goto err;

	ret = nvl_set(qso->additional, "wsjtx-raw-udp",
		      VAL_ALLOC_BLOB_STATIC(raw, rawlen));
	if (ret)
		goto err;

	/*
	 * Now, save the QSO.
	 */
	ret = qso_save(qso, AT_FDCWD);
	if (!ret) {
		struct str *mode, *call, *grid;

		mode = qso_get_mode(&qso->tx);
		call = qso_get_station_call(&qso->rx);
		grid = qso_get_grid(&qso->rx);

		cmn_err(CE_INFO, "wsjtx: saved %s %s %s %s",
			qso->uuid_str, str_cstr(mode), str_cstr(call),
			str_cstr(grid));

		announce_notify_qso(ANNOUNCE_QSO_NEW, &qso->uuid);

		str_putref(mode);
		str_putref(call);
		str_putref(grid);
	}

err:
	qso_putref(qso);

	return ret;
}

static void wsjtx_process(void *buf, size_t len, void *unused)
{
	struct rawmessage *raw = buf;
	struct message msg;
	struct buffer payload;
	int ret;

	if (len < (3 * sizeof(uint32_t))) {
		cmn_err(CE_INFO, "Message is too short (got %zu bytes, "
			"need at least %zu)", len, 3 * sizeof(uint32_t));
		return;
	}

	/*
	 * Handle the header and message type
	 */
	msg.magic = be32_to_cpu(raw->magic);
	msg.schema = be32_to_cpu(raw->schema);
	msg.type = be32_to_cpu(raw->type);

	if (msg.magic != WSJTX_MAGIC) {
		cmn_err(CE_INFO, "Magic mismatch (got %08x, expected %08x)",
			msg.magic, WSJTX_MAGIC);
		return;
	}

	if (msg.schema != WSJTX_SCHEMA) {
		cmn_err(CE_INFO, "Unsupported schema (got %u, expected %u)",
			msg.schema, WSJTX_SCHEMA);
		return;
	}

	buffer_init_static(&payload, &raw->payload,
			   len - offsetof(struct rawmessage, payload),
			   len - offsetof(struct rawmessage, payload),
			   false);

	switch (msg.type) {
		case 5: /* QSO Logged */
			if (!decode_qso_logged(&msg, &payload))
				return;

			ret = log_qso(&msg, buf, len);
			if (ret)
				cmn_err(CE_WARN, "Failed to log QSO: %s",
					xstrerror(ret));
			break;
		default:
			/* ignore */
			break;
	}
}

int wsjtx_start(void)
{
	int ret;

	ret = udpsvc(WSJTX_HOST, WSJTX_PORT, wsjtx_process, NULL);
	if (ret)
		cmn_err(CE_ERROR, "wsjtx: failed to start listener: %s",
			xstrerror(ret));

	return ret;
}
